#!/usr/bin/env python2
# Paper Wallet for BlackCoin

"""

    %prog [options]

%prog - A Python Paper Wallet Generator
==============================================

%prog generates a (encrypted) PDF file that displays the private key and 
public *Coin address.

For the key generation the script http://we.lovebitco.in/paperwal.py is employed.

---------------------------------------------------------------------------------
The author of mondero takes no responsibility for the cryptographic correctnes
of the generated keys.
---------------------------------------------------------------------------------
"""

epilog = """
Examples:
./monedero.py
    
    Query for a PDF passphrase and requests entropy key presses
    and generates wallet.pdf
    
./monedero.py -o my_wallet.pdf 

    Write the file my_wallet.pdf
    
./monedero.py -P mysupersecurepasswd

    For Batch Generation of several paper wallets, you can provide 
    the PDF passphrase by command line. !Attention! the passphrase 
    might be stored in the command line history.

See also http://monedero.readthedocs.io for full documentation.

"""
import monedero.KeyGenerator
import monedero.PdfWallet



import readline, rlcompleter
from optparse import OptionParser, OptionGroup
import getpass
from reportlab.lib.pagesizes import A4

readline.parse_and_bind("tab:complete")

class MyOptionParser(OptionParser):
    def format_epilog(self, formatter):
        return self.epilog

def getOptions():
    parser = MyOptionParser( usage = __doc__,
                           version = "%prog 1.0",
                           epilog = epilog)
    parser.add_option("-e", "--encrypted", action='store_true', dest="encrypted", default=False,
                      help="create BIP38-encrypted privkey (takes a LONG time)")
    parser.add_option("-v", "--validate", action='store_true', dest="validate", default=False,
                      help="enable extensive system tests for entropy")
    parser.add_option("-s", "--silent", action='store_true', dest="silent", default=False,
                      help="disable most console output except address")
    parser.add_option("-l", "--loop", action='store_true', dest="repeat", default=False,
                      help="restart instead of exit")
    parser.add_option("-p", "--nopause", action='store_true', dest="nopause", default=True,
                      help="disable the pause before exiting")
    parser.add_option("-d", "--doublecalc", action='store_true', dest="doublecalc", default=False,
                      help="calculate twice and test results")
    parser.add_option("-z", "", dest='just a helpful hint', default='',
                      help="try ctrl-tab to abort the program")
    parser.add_option("-o", "--output", dest='pdffilename', default='wallet.pdf',
                      help="name of the generated PDF file (default: wallet.pdf)")
    parser.add_option("-P", "--password", dest='pdfpasswd', default=None,
                      help="protect PDF file with password")


    entropy_warning = OptionGroup(parser, "Warning",
                                          "If you use this option, you should supply REAL randomly generated entropy. "
                                          "It is probably a good idea not to reuse a seed.")
    entropy_warning.add_option("-r", "--entropy", dest='entropy', default='',
                               help="random seed instead of keypresses, 64+ characters")
    parser.add_option_group(entropy_warning)
    (options, args) = parser.parse_args()
    return (options, args)

def validateOptions(options):
    if options.pdfpasswd == None:
        p1 = getpass.getpass("enter pdf password:").rstrip()
        if len(p1):
            p2 = getpass.getpass("confirm pdf password:").rstrip()
            if p1 == p2:
                options.pdfpasswd = p1
            else:
                raise Exception("invalid password")
        else:
            options.pdfpasswd = ""
            
    if len(options.pdfpasswd) < 1:
        print "PdfFile will not be encrypted!"
        options.pdfpasswd = None



if __name__ == "__main__":
    (options, args) = getOptions()
    validateOptions(options)

    privk, pubk = monedero.KeyGenerator.paperwal(options, args)

    w = monedero.PdfWallet.PdfWallet(privk, pubk,
                                fileName = options.pdffilename,
                                encryptPdf = options.pdfpasswd,
                                pageSize=A4)
    w.render()
    w.save()
    print "Wrote file: ", options.pdffilename
    
